from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    overload,
)

from ._types import InteractionT

if TYPE_CHECKING:
    from collections.abc import Callable

    from discord import (
        AutoShardedClient,
        Client,
    )
    from discord.app_commands import Command, ContextMenu
    from discord.ext.commands import (  # pyright: ignore[reportMissingTypeStubs]
        AutoShardedBot,
        Bot,
    )

    class WaitForOverloads(Generic[InteractionT]):
        @overload
        async def wait_for(  # pyright: ignore[reportNoOverloadImplementation]
            self: Client | AutoShardedClient | Bot | AutoShardedBot | WaitForOverloads,
            event: Literal["app_command_completion"],
            /,
            *,
            check: (
                Callable[[InteractionT, Command[Any, ..., Any] | ContextMenu], bool]
                | None
            ) = ...,
            timeout: float | None = ...,
        ) -> tuple[InteractionT, Command[Any, ..., Any] | ContextMenu]: ...

        @overload
        async def wait_for(
            self: Client | AutoShardedClient | Bot | AutoShardedBot | WaitForOverloads,
            event: Literal["interaction"],
            /,
            *,
            check: Callable[[InteractionT], bool] | None = ...,
            timeout: float | None = ...,
        ) -> InteractionT: ...

else:

    class WaitForOverloads(Generic[InteractionT]):
        pass



from typing import Any, overload  # noqa: I002
from collections.abc import Sequence

import discord
from discord.utils import MISSING

__all__ = ("CustomGuildInteraction", "CustomInteraction")


class CustomInteraction(discord.Interaction):
    @property
    def author(self) -> discord.Member | discord.User:
        return self.user

    @property
    def voice_client(self) -> discord.VoiceProtocol | None:
        return self.guild and self.guild.voice_client

    @overload
    async def send(
        self,
        content: str = MISSING,
        *,
        username: str = MISSING,
        ephemeral: bool = MISSING,
        file: discord.File = MISSING,
        files: Sequence[discord.File] = MISSING,
        embed: discord.Embed = MISSING,
        embeds: Sequence[discord.Embed] = MISSING,
        allowed_mentions: discord.AllowedMentions = MISSING,
        view: discord.ui.View = MISSING,
        thread: discord.abc.Snowflake = MISSING,
        thread_name: str = MISSING,
        suppress_embeds: bool = MISSING,
        silent: bool = MISSING,
        applied_tags: list[discord.ForumTag] = MISSING,
        poll: discord.Poll = MISSING,
    ) -> discord.WebhookMessage: ...

    @overload
    async def send(
        self,
        content: str = MISSING,
        *,
        username: str = MISSING,
        ephemeral: bool = MISSING,
        file: discord.File = MISSING,
        files: Sequence[discord.File] = MISSING,
        embed: discord.Embed = MISSING,
        embeds: Sequence[discord.Embed] = MISSING,
        allowed_mentions: discord.AllowedMentions = MISSING,
        view: discord.ui.View = MISSING,
        thread: discord.abc.Snowflake = MISSING,
        thread_name: str = MISSING,
        suppress_embeds: bool = MISSING,
        silent: bool = MISSING,
        applied_tags: list[discord.ForumTag] = MISSING,
        poll: discord.Poll = MISSING,
    ) -> None: ...

    async def send(
        self,
        content: str = MISSING,
        *,
        username: str = MISSING,
        ephemeral: bool = False,
        file: discord.File = MISSING,
        files: Sequence[discord.File] = MISSING,
        embed: discord.Embed = MISSING,
        embeds: Sequence[discord.Embed] = MISSING,
        allowed_mentions: discord.AllowedMentions = MISSING,
        view: discord.ui.View = MISSING,
        thread: discord.abc.Snowflake = MISSING,
        thread_name: str = MISSING,
        suppress_embeds: bool = False,
        silent: bool = False,
        applied_tags: list[discord.ForumTag] = MISSING,
        poll: discord.Poll = MISSING,
    ) -> discord.WebhookMessage | None:
        send_kwargs: dict[str, Any] = {
            "content": content,
            "username": username,
            "ephemeral": ephemeral,
            "file": file,
            "files": files,
            "embed": embed,
            "embeds": embeds,
            "allowed_mentions": allowed_mentions,
            "view": view,
            "thread": thread,
            "thread_name": thread_name,
            "suppress_embeds": suppress_embeds,
            "silent": silent,
            "applied_tags": applied_tags,
            "poll": poll,
        }

        if self.response.is_done():
            return await self.followup.send(wait=True, **send_kwargs)

        return await self.response.send_message(**send_kwargs)


class CustomGuildInteraction(CustomInteraction):
    author: discord.Member  # pyright: ignore[reportIncompatibleMethodOverride]
    user: discord.Member  # pyright: ignore[reportIncompatibleVariableOverride]
    guild: discord.Guild  # pyright: ignore[reportIncompatibleMethodOverride]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Concatenate, ParamSpec, TypeVar
from collections.abc import Callable, Coroutine

import discord
from discord.ext.commands import Cog  # pyright: ignore[reportMissingTypeStubs]

if TYPE_CHECKING:
    from typing_extensions import TypeVar  # noqa: TC004

    from discord import Client, Interaction

    ClientT_co = TypeVar("ClientT_co", bound=Client, covariant=True, default=Client)
    InteractionT = TypeVar(
        "InteractionT",
        bound=Interaction,
        default=Interaction,
    )
else:
    ClientT_co = TypeVar("ClientT_co", bound="Client", covariant=True)
    InteractionT = TypeVar("InteractionT", bound="Interaction")


CoroT = TypeVar("CoroT")
Coro = Coroutine[Any, Any, CoroT]


P = ParamSpec("P")
T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])

Binding = discord.app_commands.Group | Cog
GroupT = TypeVar("GroupT", bound=Binding)
UnboundError = Callable[
    [InteractionT, discord.app_commands.AppCommandError],
    Coro[Any],
]
Error = (
    Callable[[GroupT, InteractionT, discord.app_commands.AppCommandError], Coro[Any]]
    | UnboundError
)
Check = Callable[[InteractionT], bool | Coro[bool]]

CommandCallback = (  # pyright: ignore[reportGeneralTypeIssues]
    Callable[Concatenate[GroupT, InteractionT, P], Coro[T]]
    | Callable[Concatenate[InteractionT, P], Coro[T]]
)

ContextMenuCallback = (
    Callable[[InteractionT, discord.Member], Coro[Any]]
    | Callable[[InteractionT, discord.User], Coro[Any]]
    | Callable[[InteractionT, discord.Message], Coro[Any]]
    | Callable[[InteractionT, discord.Member | discord.User], Coro[Any]]
)
ChoiceT = TypeVar("ChoiceT", str, int, float, str | int | float)

AutocompleteCallback = (  # pyright: ignore[reportGeneralTypeIssues]
    Callable[
        [InteractionT, GroupT, str],
        Coro[list[discord.app_commands.Choice[ChoiceT]]],
    ]
    | Callable[[InteractionT, str], Coro[list[discord.app_commands.Choice[ChoiceT]]]]
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, TypeVar

import discord
from discord.ext import commands
import discord.state as ds

from ._types import (
    ClientT_co,
    InteractionT,
)
from ._wait_for_overloads import WaitForOverloads

__all__ = ("InteractionClient",)

if TYPE_CHECKING:
    from typing_extensions import Self, TypeVar  # noqa: TC004

    InteractionT = TypeVar(
        "InteractionT",
        bound=discord.Interaction,
        default=discord.Interaction,
    )

    from discord.types.gateway import (
        InteractionCreateEvent as InteractionCreateEventPayload,
    )
else:
    InteractionT = TypeVar("InteractionT", bound=discord.Interaction)


if TYPE_CHECKING:
    from collections.abc import Callable

    from discord.ext.commands.bot import (  # pyright: ignore[reportMissingTypeStubs]
        PrefixType,
    )

    from ._types import (
        AutocompleteCallback,
        Check,
        ChoiceT,
        CommandCallback,
        ContextMenuCallback,
        GroupT,
        P,
        T,
    )

# fmt: off
__all__ = (  # noqa: RUF022
    "InteractionClient",
    "InteractionAutoShardedClient",
    "InteractionBot",
    "InteractionAutoShardedBot",
    "command",
    "context_menu",
    "check",
    "autocomplete",
)
# fmt: on


class _CustomInteractionState(
    ds.ConnectionState[ClientT_co],
    Generic[ClientT_co, InteractionT],
):
    def __init__(
        self,
        *args: Any,
        interaction_cls: type[InteractionT],
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.__interaction_cls: type[InteractionT] = interaction_cls

    def parse_interaction_create(self, data: InteractionCreateEventPayload) -> None:
        interaction = self.__interaction_cls(data=data, state=self)

        if (
            data["type"] in (2, 4) and self._command_tree
        ):  # application command and auto complete
            self._command_tree._from_interaction(interaction)
        elif data["type"] == 3:  # interaction component
            # These keys are always there for this interaction type
            inner_data = data["data"]
            custom_id = inner_data["custom_id"]
            component_type = inner_data["component_type"]
            self._view_store.dispatch_view(component_type, custom_id, interaction)
        elif data["type"] == 5:  # modal submit
            # These keys are always there for this interaction type
            inner_data = data["data"]
            custom_id = inner_data["custom_id"]
            components = inner_data["components"]
            self._view_store.dispatch_modal(custom_id, interaction, components)

        self.dispatch("interaction", interaction)


class InteractionClient(
    WaitForOverloads[InteractionT],
    discord.Client,
):
    def __init__(
        self,
        *args: Any,
        intents: discord.Intents,
        interaction_cls: type[InteractionT],
        **kwargs: Any,
    ) -> None:
        # passed to _CustomInteractionState via _get_state
        kwargs["interaction_cls"] = interaction_cls
        super().__init__(*args, intents=intents, **kwargs)

        error_msg = f"interaction_cls must be a subclass of discord.Interaction not {interaction_cls!r}"
        try:
            if not issubclass(interaction_cls, discord.Interaction):
                raise TypeError(error_msg)
        except TypeError as exc:
            # inspect.isclass? no thanks.
            if "must be a class" in str(exc):
                raise TypeError(error_msg) from None

            raise

    if not TYPE_CHECKING:

        def _get_state(self, **options: Any) -> _CustomInteractionState[Self]:
            return _CustomInteractionState(
                dispatch=self.dispatch,
                handlers=self._handlers,
                hooks=self._hooks,
                http=self.http,
                **options,
            )

    if TYPE_CHECKING:

        async def on_interaction(self, interaction: InteractionT) -> Any:
            pass


class InteractionAutoShardedClient(
    InteractionClient[InteractionT],
    discord.AutoShardedClient,
):
    pass


class InteractionBot(
    commands.bot.BotBase,
    InteractionClient[InteractionT],
):
    if TYPE_CHECKING:
        from .tree import InteractionTree  # noqa: PLC0415

        tree: InteractionTree[  # pyright: ignore[reportIncompatibleMethodOverride]
            InteractionT
        ]

        def __init__(
            self,
            command_prefix: PrefixType[commands.Bot],
            *args: Any,
            intents: discord.Intents,
            interaction_cls: type[InteractionT],
            **kwargs: Any,
        ) -> None: ...


class InteractionAutoShardedBot(
    InteractionBot[InteractionT],
    InteractionAutoShardedClient[InteractionT],
):
    pass


def command(  # pyright: ignore[reportGeneralTypeIssues]
    *,
    name: str | discord.app_commands.locale_str = discord.utils.MISSING,
    description: str | discord.app_commands.locale_str = discord.utils.MISSING,
    nsfw: bool = False,
    auto_locale_strings: bool = True,
    extras: dict[Any, Any] = discord.utils.MISSING,
) -> Callable[
    [CommandCallback[GroupT, InteractionT, P, T]],
    discord.app_commands.Command[GroupT, P, T],
]:
    return discord.app_commands.command(  # pyright: ignore[reportReturnType]
        name=name,
        description=description,
        nsfw=nsfw,
        auto_locale_strings=auto_locale_strings,
        extras=extras,
    )


def context_menu(  # pyright: ignore[reportGeneralTypeIssues]
    *,
    name: str | discord.app_commands.locale_str = discord.utils.MISSING,
    nsfw: bool = False,
    auto_locale_strings: bool = True,
    extras: dict[Any, Any] = discord.utils.MISSING,
) -> Callable[[ContextMenuCallback], discord.app_commands.ContextMenu]:
    return discord.app_commands.context_menu(
        name=name,
        nsfw=nsfw,
        auto_locale_strings=auto_locale_strings,
        extras=extras,
    )


def check(predicate: Check) -> Callable[[T], T]:
    return discord.app_commands.check(predicate)


def autocomplete(  # pyright: ignore[reportGeneralTypeIssues]
    **parameters: AutocompleteCallback[InteractionT, GroupT, ChoiceT],
) -> Callable[[T], T]:
    return discord.app_commands.autocomplete(
        **parameters,  # pyright: ignore[reportArgumentType]
    )



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
)

import discord

from ._types import InteractionT

if TYPE_CHECKING:
    from typing_extensions import Self


__all__ = ("InteractionTree",)


class InteractionTree(
    discord.app_commands.CommandTree,
    Generic[InteractionT],
):

    if TYPE_CHECKING:
        from collections.abc import Callable, Sequence  # noqa: PLC0415

        from ._types import (  # noqa: PLC0415
            CommandCallback,
            ContextMenuCallback,
            Coro,
            P,
            T,
        )

        def command(
            self: Self,
            *,
            name: str | discord.app_commands.locale_str = discord.utils.MISSING,
            description: str | discord.app_commands.locale_str = discord.utils.MISSING,
            nsfw: bool = False,
            guild: discord.abc.Snowflake | None = discord.utils.MISSING,
            guilds: Sequence[discord.abc.Snowflake] = discord.utils.MISSING,
            auto_locale_strings: bool = True,
            extras: dict[Any, Any] = discord.utils.MISSING,
        ) -> Callable[
            [CommandCallback[discord.app_commands.Group, InteractionT, P, T]],
            discord.app_commands.Command[discord.app_commands.Group, P, T],
        ]: ...

        def context_menu(
            self: Self,
            *,
            name: str | discord.app_commands.locale_str = discord.utils.MISSING,
            nsfw: bool = False,
            guild: discord.abc.Snowflake | None = discord.utils.MISSING,
            guilds: Sequence[discord.abc.Snowflake] = discord.utils.MISSING,
            auto_locale_strings: bool = True,
            extras: dict[Any, Any] = discord.utils.MISSING,
        ) -> Callable[[ContextMenuCallback], discord.app_commands.ContextMenu]: ...

        async def interaction_check(  # pyright: ignore[reportIncompatibleMethodOverride]
            self,
            interaction: InteractionT,
            /,
        ) -> bool: ...

        def error(  # pyright: ignore[reportIncompatibleMethodOverride]
            self,
            coro: Callable[
                [InteractionT, discord.app_commands.AppCommandError],
                Coro[Any],
            ],
        ) -> Callable[
            [InteractionT, discord.app_commands.AppCommandError],
            Coro[Any],
        ]: ...

        async def on_error(  # pyright: ignore[reportIncompatibleMethodOverride]
            self,
            interaction: InteractionT,
            error: discord.app_commands.AppCommandError,
            /,
        ) -> None: ...



"""An extension for discord.py that allows you to subclass `discord.Interaction` to add custom functionality."""  # noqa: I002

from .core import *
from .interactions import *
from .tree import *

__author__ = "Soheab"
__version__ = "0.0.1"
__license__ = "MPL-2.0"
